{
  "nbformat": 4,
  "nbformat_minor": 0,
  "metadata": {
    "colab": {
      "provenance": []
    },
    "kernelspec": {
      "name": "python3",
      "display_name": "Python 3"
    },
    "language_info": {
      "name": "python"
    }
  },
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "id": "_udrXQdbyRYd"
      },
      "outputs": [],
      "source": [
        "from tqdm import tqdm\n",
        "import numpy as np\n",
        "from matplotlib import pyplot as plt\n",
        "\n",
        "class Camera():\n",
        "  def __init__(self, position, viewport_size, viewport_distance):\n",
        "    self.position = position\n",
        "    self.viewport_size = viewport_size\n",
        "    self.viewport_distance = viewport_distance\n",
        "\n",
        "class Light():\n",
        "  def __init__(self, eidos, intensity, position = None, direction = None):\n",
        "    assert(eidos in ['ambient', 'point', 'directional'])\n",
        "    assert(intensity >= 0.0 and intensity <= 1.0)\n",
        "    self.eidos = eidos\n",
        "    self.intensity = intensity\n",
        "    if(eidos == 'ambient'):\n",
        "      assert(position is None)\n",
        "      assert(direction is None)\n",
        "    elif(eidos == 'point'):\n",
        "      assert(position is not None)\n",
        "      assert(direction is None)\n",
        "    elif(eidos == 'directional'):\n",
        "      assert(position is None)\n",
        "      assert(direction is not None)\n",
        "    self.position = position\n",
        "    self.direction = direction\n",
        "\n",
        "class Sphere():\n",
        "  def __init__(self, center, radius, color, specularity = None):\n",
        "    assert(radius > 0)\n",
        "    self.center = center\n",
        "    self.radius = radius\n",
        "    self.color = color\n",
        "    self.specularity = specularity  # default τιμή = None (δηλαδή δεν θέλω να είναι specular/γυαλιστερό το αντικείμενο)\n",
        "\n",
        "class Canvas():   # (σύμβαση: το όνομα του αντικειμένου θα αρχίζει με κεφαλαίο)\n",
        "  def __init__(self, Ch, Cw):\n",
        "    # Αυτή η συνάρτηση είναι ο \"constructor\".\n",
        "    # Τρέχει αυτόματα με το που δημιουργείται ένα αντικείμενο τύπου \"Canvas\".\n",
        "    self.Ch = Ch\n",
        "    self.Cw = Cw\n",
        "    self.data = np.zeros([Ch, Cw, 3])\n",
        "    print(f'Δημιουργήθηκε ένας καμβάς με μέγεθος {Ch}x{Cw}')\n",
        "\n",
        "  def putPixel(self, Cx, Cy, value):\n",
        "    # Αυτή η συνάρτηση θα \"βάψει\" μία συγκεκριμένη θέση (Cx, Cy)\n",
        "    # στον καμβά (canvas), χρησιμοποιώντας την τιμή \"value\".\n",
        "    # Η τιμή value είναι ακέραιος στο 0..255 .\n",
        "    #\n",
        "    # Input.\n",
        "    #   canvas      Πίνακας numpy μεγέθους Ch x Cw.\n",
        "    #   Cx,Cy       Συντεταγμένες εικονοστοιχείου στο ΣΣ του καμβά. Είναι η θέση που θέλω να βάψω.\n",
        "    #   value       To χρώμα με το οποίο θα βάψω την παραπάνω θέση.\n",
        "    #\n",
        "    # Output.\n",
        "    #   canvas      Ο τροποποιημένος καμβάς.\n",
        "    #\n",
        "    Ch, Cw, _ = self.data.shape\n",
        "    Si = Ch // 2 - Cy\n",
        "    Sj = Cw // 2 + Cx\n",
        "    if(Si >= 0 and Si < Ch and Sj >= 0 and Sj < Cw):\n",
        "      self.data[Si, Sj, :] = np.uint8(value)\n",
        "\n",
        "  def show(self):\n",
        "    plt.imshow(np.uint8(self.data))\n",
        "    plt.show()\n",
        "\n",
        "\n",
        "\n",
        "BACKGROUND_COLOR = np.array([0, 0, 0])\n",
        "sphere_list = [\n",
        "    Sphere(center = np.array([0, -1, 3]), radius = 1, color = np.array([255, 0, 0]), specularity = 500),     # Κόκκινη\n",
        "    Sphere(center = np.array([2,  0, 4]), radius = 1, color = np.array([0, 0, 255]), specularity = 500),     # Μπλε\n",
        "    Sphere(center = np.array([-2, 0, 4]), radius = 1, color = np.array([0, 255, 0]), specularity = 10),     # Πράσινη\n",
        "    Sphere(center = np.array([0, -5001, 0]), radius = 5000, color = np.array([255, 255, 0]), specularity = 1000),\n",
        "]\n",
        "\n",
        "light_list = [\n",
        "    Light(eidos = 'ambient', intensity = 0.0),\n",
        "    Light(eidos = 'point', intensity = 0.6, position = np.array([2, 1, 0])),\n",
        "    Light(eidos = 'directional', intensity = 0.2, direction = np.array([1, 4, 4]))\n",
        "]\n",
        "\n",
        "\n",
        "def canvasToViewport(Cx, Cy, Vw, Vh, Cw, Ch, viewport_distance):\n",
        "  Vx = Cx*(Vw/Cw)\n",
        "  Vy = Cy*(Vh/Ch)\n",
        "  Vz = viewport_distance\n",
        "  return(np.array([Vx, Vy, Vz]))\n",
        "\n",
        "\n",
        "def intersectRaySphere(O, D, sphere):\n",
        "  ###\n",
        "  # Έλεγχος τομής ακτίνας με αρχικά-τελικά σημεία ΟD και σφαίρας \"sphere\"\n",
        "  # Επιστρέφει t1, t2 (συντελεστές που πρέπει να πολλαπλασιάσουν το OD ώστε να πάρουμε\n",
        "  # την τομή ως τελικό σημείο)\n",
        "  #\n",
        "  # Αν δεν υπάρχει τομή, επιστρέφω t1, t2 = +np.inf, +np.inf\n",
        "  #\n",
        "  C = sphere.center\n",
        "  r = sphere.radius\n",
        "  #color = sphere.color\n",
        "  CO = O - C\n",
        "  # Λύση δευτεροβάθμιας εξίσωσης.\n",
        "  a = np.dot(D, D)\n",
        "  b = 2*np.dot(CO, D)\n",
        "  c = np.dot(CO, CO) - r**2\n",
        "  # Έλεγχος αν η ακτίνα χτυπάει το αντικείμενο\n",
        "  diakrinousa = b**2 - 4*a*c\n",
        "  if(diakrinousa < 0):\n",
        "    return(+np.inf, +np.inf)\n",
        "  t1 = (-b - np.sqrt(diakrinousa))/(2*a)\n",
        "  t2 = (-b + np.sqrt(diakrinousa))/(2*a)\n",
        "  return t1, t2\n",
        "\n",
        "def computeLighting(P, N, V, s):\n",
        "  # Υπολογίζει σαν αποτέλεσμα ένα συντελεστή\n",
        "  # που πρέπει να πολλαπλασιαστεί με μια χρωματική τιμή.\n",
        "  #\n",
        "  # Είσοδος :\n",
        "  #       P       Συντεταγμένες του σημείου που με ενδιαφέρει (το σημείο που \"χτυπάει\" η ακτίνα που εξέπεμψα)\n",
        "  #       N       Κανονικό διάνυσμα (σημείο κάθετο στην επιφάνεια του αντικειμένου που \"χτύπησε\" η ακτίνα)\n",
        "  #       V       View vector (διάνυσμα με αρχικό σημείο το σημείο που με ενδιαφέρει και τελικό σημείο τον παρατηρητή)\n",
        "  #       s       Όρος που ρυθμίζει το πόσο γυαλιστερό είναι το αντικείμενο (specularity).\n",
        "  #               Αν είναι None, θεωρώ ότι είναι matte (άρα όχι γυαλιστερό)\n",
        "  #\n",
        "  I = 0.0\n",
        "  for light in light_list:\n",
        "    if(light.eidos == 'ambient'):\n",
        "      I += light.intensity\n",
        "    else:\n",
        "      if(light.eidos == 'point'):\n",
        "        L = light.position - P\n",
        "        t_max = 1.0\n",
        "      elif(light.eidos == 'directional'):\n",
        "        L = light.direction\n",
        "        t_max = +np.inf\n",
        "      ######################## END IF\n",
        "\n",
        "      ## Όρος diffuse\n",
        "      n_dot_l = np.dot(N, L)\n",
        "      if(n_dot_l > 0):\n",
        "        I += light.intensity * n_dot_l / (np.linalg.norm(L) * np.linalg.norm(N))\n",
        "\n",
        "      ## Όρος specular\n",
        "      if(s is not None):\n",
        "        R = 2*N*np.dot(N, L) - L\n",
        "        r_dot_v = np.dot(R, V)\n",
        "        if(r_dot_v > 0):\n",
        "          I += light.intensity * (r_dot_v / (np.linalg.norm(R) * np.linalg.norm(V)))**s\n",
        "  ############## END FOR\n",
        "  if(I > 1.0):\n",
        "    I = 1.0\n",
        "  return(I)\n",
        "\n",
        "def closestIntersection(O, D, t_min, t_max):\n",
        "  nearest_t = +np.inf     # Το μέχρι στιγμής μικρότερο t (δηλαδή αντιστοιχεί στην κοντινότερη, αν υπάρχει, τομή)\n",
        "  nearest_sphere = None\n",
        "  for sphere in sphere_list:\n",
        "    t1, t2 = intersectRaySphere(O, D, sphere)\n",
        "    if(t1 > t_min and t1 < t_max and t1 < nearest_t):\n",
        "      nearest_t = t1\n",
        "      nearest_sphere = sphere\n",
        "    if(t2 > t_min and t2 < t_max and t2 < nearest_t):\n",
        "      nearest_t = t2\n",
        "      nearest_sphere = sphere\n",
        "  ######## end of for loop\n",
        "  return(nearest_sphere, nearest_t)\n",
        "\n",
        "def traceRay(O, D, t_min, t_max):\n",
        "  nearest_sphere, nearest_t = closestIntersection(O, D, t_min, t_max)\n",
        "  if(nearest_sphere is None):\n",
        "    return(BACKGROUND_COLOR)\n",
        "  else:\n",
        "    P = O + nearest_t*D\n",
        "    N = P - nearest_sphere.center\n",
        "    syntelestis = computeLighting(P, N, -D, nearest_sphere.specularity)\n",
        "    color = np.uint8(nearest_sphere.color * syntelestis)\n",
        "    return(color)\n",
        "\n",
        "\n",
        "# Δημιούργησε τον καμβά.\n",
        "mycanvas = Canvas(Ch = 300, Cw = 300)\n",
        "# Τοποθέτησε την κάμερα.\n",
        "mycamera = Camera(position = np.array([0, 0, 0]), viewport_size = np.array([1.0, 1.0]), viewport_distance=1)\n",
        "for x in tqdm(range(-mycanvas.Cw//2, +mycanvas.Cw//2)):\n",
        "  for y in range(-mycanvas.Ch//2, +mycanvas.Ch//2):\n",
        "    # Μετατροπή συντατεγμένων καμβά (Cx,Cy) σε συντεταγμένες στον 3Δ χώρο (VX,VY,VZ)\n",
        "    D = canvasToViewport(Cx = x, Cy = y, Vw = mycamera.viewport_size[0], Vh = mycamera.viewport_size[1], Cw = mycanvas.Cw, Ch = mycanvas.Ch, viewport_distance=mycamera.viewport_distance)\n",
        "    # Παρακολούθηση ακτίνας (αρχικό σημείο θέση της κάμερας, τελικό σημείο VΧ,VΥ,VΖ που βρήκα προηγουμένως). Αποτέλεσμα ένα χρώμα\n",
        "    color = traceRay(mycamera.position, D, t_min = 1, t_max = +np.inf)\n",
        "    # Βάψε τον καμβά (θέση Cx,Cy) με το χρώμα.\n",
        "    mycanvas.putPixel(Cx=x, Cy=y, value=color)\n",
        "\n",
        "\n",
        "mycanvas.show()"
      ]
    }
  ]
}